#[cfg(windows)]
extern crate winapi;

//use winapi::um::winnt::SystemPowerInformation;
use std::mem::size_of; // buffer/struct size
use std::ptr::null_mut; // NULL
use std::time::Duration;
use winapi::shared::minwindef::{DWORD, PULONG, UCHAR, ULONG, HMODULE};
use winapi::um::errhandlingapi::GetLastError;
use winapi::um::handleapi::CloseHandle;
use winapi::um::processthreadsapi::{GetCurrentProcess, OpenProcess};
use winapi::um::psapi::{K32EnumProcessModulesEx, K32EnumProcesses, K32GetModuleBaseNameW};
use winapi::um::systemtopologyapi::GetNumaHighestNodeNumber;
use winapi::um::winbase::{GetProcessAffinityMask, SetProcessAffinityMask};
use winapi::um::winnt::{HANDLE, WCHAR};

use std::mem::MaybeUninit;

fn main() {
    //let mut state = State::new();
    let cpu_info = get_cpu_info();
    println!("Hello, {:?}!", cpu_info);

    // https://docs.microsoft.com/en-us/windows/win32/procthread/numa-support

    //let cores_to_use = vec![CoreId(0), CoreId(2), CoreId(4), CoreId(6)];
    let cores_to_use = vec![CoreId(0), CoreId(1), CoreId(2), CoreId(3)];
    set_process_affinity("*", &cores_to_use);

    let three_secs = Duration::from_secs(3);

    /* Demo the switching
    let proc_handle = unsafe { GetCurrentProcess() };
    for i in 0..cpu_info.unwrap() {
        unsafe { SetProcessAffinityMask(proc_handle, (1 << i) as u32) };
        make_load(three_secs * 3);
    }
    */

    std::thread::sleep(three_secs);
}

fn get_cpu_info() -> Result<usize, &'static str> {
    let mut numa_node_count: ULONG = 0xffffffff;
    // https://docs.microsoft.com/en-us/windows/win32/api/systemtopologyapi/nf-systemtopologyapi-getnumahighestnodenumber
    let r = unsafe { GetNumaHighestNodeNumber(&mut numa_node_count) };
    if r == 0 {
        println!("Error: {}", os_get_last_error());
        println!(
            "(DD) GetNumaHighestNodeNumber({}) -> {}",
            numa_node_count, r
        );
        return Err("unable to get GetNumaHighestNodeNumber");
    }

    if numa_node_count > 0 {
        //Ok(numa_node_count);
        unimplemented!("NUMA Systems not implemented yet!");
    } else {
        let proc_handle = unsafe { GetCurrentProcess() };
        let mut lpProcessAffinityMask: usize = 0b0010_0000;
        //unsafe { SetProcessAffinityMask(proc_handle, lpProcessAffinityMask as u32) };
        let mut lpSystemAffinityMask: usize = usize::max_value();
        let r = unsafe {
            GetProcessAffinityMask(
                proc_handle,
                &mut lpProcessAffinityMask as *mut _ as *mut _,
                &mut lpSystemAffinityMask as *mut _ as *mut _,
            )
        };
        if r == 0 {
            return Err("unable to get GetProcessAffinityMask()");
        }
        Ok(lpSystemAffinityMask.count_ones() as usize)
    }
}

fn os_get_last_error() -> String {
    use winapi::shared::ntdef::{LANG_NEUTRAL, SUBLANG_DEFAULT, MAKELANGID, LPWSTR};
    use winapi::um::winbase::*;
    let code = unsafe { GetLastError() };

    let mut str_ptr: LPWSTR = null_mut();
    let dword_len = unsafe { winapi::um::winbase::FormatMessageW(
        FORMAT_MESSAGE_ALLOCATE_BUFFER | 
            FORMAT_MESSAGE_FROM_SYSTEM |
            FORMAT_MESSAGE_IGNORE_INSERTS,
        null_mut(),
        code,
        MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT) as u32,
        &mut str_ptr as *mut _ as * mut _,
        0, 
        null_mut()
    ) };
    let msg_slice = unsafe {
        std::slice::from_raw_parts(str_ptr, dword_len as usize)
    };

    let r = format!("Code: {}; Message: {}", code, String::from_utf16_lossy(msg_slice));

    unsafe { winapi::um::winbase::LocalFree(str_ptr as * mut _) };

    r
}

fn make_load(timeout: Duration) {
    let start = std::time::Instant::now();
    let mut i = 0usize;
    while start.elapsed() < timeout {
        if i < 1024 {
            i += 1;
        } else {
            i = 0;
        }
    }
}

#[derive(Debug, PartialEq)]
pub struct CoreId(u8);

fn affinity_to_mask(l: &Vec<CoreId>) -> u32 {
    let mut mask = 0;

    for i in l {
        mask |= 1u32 << i.0
    }

    mask
}
fn mask_to_affinity(mask: u32) -> Vec<CoreId> {
    let mut r = Vec::with_capacity(mask.count_ones() as usize);

    for i in 0..32 {
        if mask & (1 << i) > 0 {
            r.push(CoreId(i));
        }
    }

    r
}

fn set_process_affinity(name: &str, cores: &Vec<CoreId>) {
    let pids = Pid::get_system_pids();
    // println!("(DD) pids {:#?}", pids);

    let handles = pids
        .unwrap()
        .into_iter()
        .map(Pid::process_handle)
        .filter_map(Result::ok)
        //.filter(|p| p.name == name)
        //.collect::<Vec<_>>()
        ;
    //println!("(DD) handles {:#?}", handles);
    for ph in handles {
        let status = if 0 == unsafe { SetProcessAffinityMask(ph.handle, affinity_to_mask(cores)) } {
            os_get_last_error()
        } else {
            "Ok".into()
        };
        println!("{}: (PID: {}) <= {}", ph.name, ph.pid.0, status);
    }
}

#[derive(Debug, Clone, PartialEq)]
pub struct Pid(u32);

#[cfg(windows)]
impl Pid {
    //use super::{Pid, K32EnumProcesses, DWORD, size_of};

    const N_PIDS: usize = 4096;

    pub fn get_system_pids() -> Result<Vec<Pid>, &'static str> {
        let mut pids = vec![Pid(0); Self::N_PIDS];
        let mut found_processes: DWORD = 0;

        // https://docs.microsoft.com/en-us/windows/win32/api/psapi/nf-psapi-enumprocesses
        if 0 == unsafe {
            K32EnumProcesses(
                pids.as_mut_ptr() as *mut u32,
                (Self::N_PIDS * size_of::<Pid>()) as u32,
                &mut found_processes,
            )
        } {
            return Err("unable to EnumProcesses");
        }
        let n_pids = found_processes as usize / size_of::<DWORD>();

        pids.truncate(n_pids);

        Ok(pids)
    }

    pub fn process_handle(self) -> Result<ProcessHandle, &'static str> {
        use winapi::um::winnt::{PROCESS_QUERY_INFORMATION, PROCESS_VM_READ, PROCESS_VM_WRITE, PROCESS_SET_INFORMATION };
        // https://docs.microsoft.com/en-us/windows/win32/procthread/process-security-and-access-rights
        let desired_access = PROCESS_VM_READ | PROCESS_QUERY_INFORMATION | PROCESS_SET_INFORMATION ;
        let inherit_handle = 0;

        // https://docs.microsoft.com/en-us/windows/win32/api/processthreadsapi/nf-processthreadsapi-openprocess
        let handle: HANDLE = unsafe { OpenProcess(desired_access, inherit_handle, self.0) };
        if handle.is_null() {
            return Err("unable to OpenProcess");
        }

        let mut module_ptr: HMODULE = null_mut();
        let mut module_dyn_needed: DWORD = 0;
        let filter_flag = winapi::um::psapi::LIST_MODULES_ALL;
        if 0 == unsafe {
            K32EnumProcessModulesEx(
                handle,
                &mut module_ptr as *mut _ as *mut _,
                size_of::<HMODULE>() as u32,
                &mut module_dyn_needed,
                filter_flag,
            )
        } {
            return Err("unable to EnumProcessModulesEx");
        }

        let mut name: Vec<WCHAR> = vec![0; winapi::shared::minwindef::MAX_PATH];
        // https://docs.microsoft.com/en-us/windows/win32/api/psapi/nf-psapi-getmodulebasenamew
        let len_name = unsafe { K32GetModuleBaseNameW(
            handle,
            module_ptr,
            name.as_mut_ptr(),
            (name.len() / size_of::<WCHAR>()) as u32,
        ) };
        if len_name == 0 {
            return Err("unable to K32GetModuleBaseNameW");
        }

        let name = String::from_utf16_lossy(&name[0..len_name as usize]);

        Ok(ProcessHandle { pid: self, handle, name })
    }
}

#[derive(Debug)]
pub struct ProcessHandle {
    pid: Pid,
    handle: HANDLE,
    name: String,
}

impl Drop for ProcessHandle {
    fn drop(&mut self) {
        unsafe { CloseHandle(self.handle) };
    }
}

/*
use  winapi::um::powerbase::CallNtPowerInformation;
fn get_cpu_info() -> ProcessorPowerInformation {
    const ProcessorInformation: usize = 11;
    let mut teststruct = MaybeUninit::<ProcessorPowerInformation>::zeroed().assume_init();
    // Fill in struct and get result code
    let result = unsafe {
        CallNtPowerInformation(
            ProcessorInformation,
            null_mut(),
            0,
            &mut teststruct,
            size_of::<ProcessorPowerInformation>() as u32
        )
    };
    // TODO check result

    teststruct
}

/// https://docs.microsoft.com/en-us/windows/win32/power/processor-power-information-str?redirectedfrom=MSDN
pub struct ProcessorPowerInformation {
    number: ULONG,
    max_mhz: ULONG,
    current_mhz: ULONG,
    mhz_limit: ULONG,
    max_idle_state: ULONG,
    current_idle_state: ULONG,
}
*/

/*
pub struct State {
    some: (),
}
impl State {
    pub fn new() -> State {
        State {
            some: (),
        }
    }
}
*/

#[cfg(test)]
mod test {
    use super::*;
    #[test]
    fn vec_to_mask() {
        let expected = 0b0001_0101;
        let result = affinity_to_mask(&vec![CoreId(2), CoreId(4), CoreId(0)]);
        assert_eq!(expected, result);
    }
    #[test]
    fn mask_to_vec() {
        let expected = vec![CoreId(0), CoreId(2), CoreId(4)];
        let result = mask_to_affinity(0b0001_0101);
        assert_eq!(expected, result);
    }
}
